package com.microsoft.office.ui.controls.widgets

import android.content.Context
import android.content.res.Configuration
import android.graphics.Point
import android.support.design.widget.BottomSheetBehavior
import android.support.design.widget.BottomSheetDialog
import android.view.FocusFinder
import android.view.Gravity
import android.view.View
import android.view.ViewGroup
import android.view.ViewGroup.LayoutParams.MATCH_PARENT
import android.view.ViewGroup.LayoutParams.WRAP_CONTENT
import android.view.accessibility.AccessibilityEvent
import android.widget.FrameLayout
import android.widget.ImageView
import android.widget.LinearLayout
import android.widget.PopupWindow

import com.microsoft.office.officespace.autogen.FSImmersiveGallerySPProxy
import com.microsoft.office.officespace.autogen.FSMenuSPProxy
import com.microsoft.office.plat.logging.Trace
import com.microsoft.office.ui.controls.callout.CustomViewProviderRegistry
import com.microsoft.office.ui.controls.colorpicker.FSColorPickerViewProvider
import com.microsoft.office.ui.controls.morecolors.MoreColorsButton
import com.microsoft.office.ui.flex.FlexDataSourceProxy
import com.microsoft.office.ui.flex.enums.AnchorTypes
import com.microsoft.office.ui.viewproviders.FSColorGridViewProvider
import com.microsoft.office.ui.viewproviders.FSComboBoxViewProvider
import com.microsoft.office.ui.viewproviders.FSImmersiveGalleryViewProvider
import com.microsoft.office.ui.viewproviders.FSMenuViewProvider
import com.microsoft.office.ui.viewproviders.IViewProvider
import com.microsoft.office.ui.viewproviders.MoreColorsViewProvider
import com.microsoft.office.ui.viewproviders.ToolboxViewProvider

import com.microsoft.office.ui.flex.R
import com.microsoft.office.ui.utils.DimensionHelper
import com.microsoft.office.ui.utils.OrientationChangeManager

import java.util.Stack
import java.util.concurrent.CopyOnWriteArrayList

class OfficeBottomSheetDialog : BottomSheetDialog, ILaunchableSurface
{
    private var mContext: Context? = null
    private var mContentView: View? = null
    private val mDismissListeners: CopyOnWriteArrayList<PopupWindow.OnDismissListener> = CopyOnWriteArrayList<PopupWindow.OnDismissListener>()
    private var mSurfaceLauncherView: ISurfaceLauncherView? = null
    private var mFocusChangeListener: View.OnFocusChangeListener
    private var mAccessibilityDelegate: View.AccessibilityDelegate
    protected var mViewProviderStack: Stack<IViewProvider>
    protected var mFlyoutBehaviorStack: Stack<FluxFlyoutBehavior>

    constructor(context: Context) : super(context, R.style.BottomSheetDialogTheme)
    {
        mContext = context
        mViewProviderStack = Stack()
        mFlyoutBehaviorStack = Stack()
        this.setOnShowListener {
            val bottomSheet = findViewById<FrameLayout>(android.support.design.R.id.design_bottom_sheet)
            val behavior = BottomSheetBehavior.from(bottomSheet)
            if ( OrientationChangeManager.getInstance().currentOrientation == Configuration.ORIENTATION_PORTRAIT)
            {
                behavior.peekHeight = (PORTRAIT_PEEK_RATIO * DimensionHelper.getActualScreenHeight()).toInt()
            }
            else
            {
                behavior.peekHeight = (LANDSCAPE_PEEK_RATIO * DimensionHelper.getActualScreenHeight()).toInt()
            }
            bottomSheet!!.requestFocus()
        }
        mFocusChangeListener = View.OnFocusChangeListener{ view, hasFocus ->
            expandBottomSheetDialog(hasFocus)
            val nextFocus = FocusFinder.getInstance().findNextFocus(view as ViewGroup, view.findFocus(), View.FOCUS_FORWARD)
            if (nextFocus != null && nextFocus != view)
            {
                nextFocus.requestFocus()
            }
        }
        mAccessibilityDelegate = object : View.AccessibilityDelegate()
        {
            override fun onRequestSendAccessibilityEvent(host: ViewGroup?, child: View?, event: AccessibilityEvent?): Boolean
            {
                if (event!!.eventType == AccessibilityEvent.TYPE_VIEW_ACCESSIBILITY_FOCUSED)
                {
                    expandBottomSheetDialog(true)
                }
                return super.onRequestSendAccessibilityEvent(host, child, event)
            }
        }
    }

    override fun dismissSurface()
    {
        dismiss()
    }

    override fun setShouldAnimate(shouldAnimate: Boolean)
    {
        // No-Op for BottomSheetDialog as animations are part of native control
    }

    override fun setControlDismissListener(onDismissListener: PopupWindow.OnDismissListener)
    {
        mDismissListeners.add(onDismissListener)
    }

    override fun removeControlDismissListener(onDismissListener: PopupWindow.OnDismissListener)
    {
        mDismissListeners.remove(onDismissListener)
    }

    override fun setHideKeyboardOnShow(hideKeyboardOnShow: Boolean)
    {
        // No-Op for BottomSheetDialog
    }

    override fun setAnchor(anchorElement: View)
    {
        // No-Op for BottomSheetDialog as we will always show it at the bottom
    }

    override fun getAnchor(): View?
    {
        // No-Op for BottomSheetDialog as we will always show it at the bottom
        return null
    }

    override fun getSurfaceLauncherView(): ISurfaceLauncherView?
    {
        return null
    }

    override fun setDataSource(dataSource: FlexDataSourceProxy, factory: IControlFactory, surfaceLauncherView: ISurfaceLauncherView, hasOverflownControls: Boolean)
    {
        mSurfaceLauncherView = surfaceLauncherView
        val viewProvider = getViewProvider(dataSource, factory, mSurfaceLauncherView)
        viewProvider.surfaceLauncherView = mSurfaceLauncherView
        viewProvider.hasOverflownControls = hasOverflownControls
        mViewProviderStack.push(viewProvider)
        mFlyoutBehaviorStack.push(FluxFlyoutBehavior(dataSource, this /*surface*/))
    }

    override fun show()
    {
        if (mContentView != null)
        {
            val container = getBottomSheetContainer()
            container.layoutParams = getLayoutParamsForContainer(mContentView)
            container.addView(mContentView)
            setContentView(container)
            mContentView = null
        }
        else
        {
            setBottomSheetContent()
        }
        super.show()
    }

    override fun dismiss()
    {
        super.dismiss()
        for (listener in mDismissListeners)
        {
            listener.onDismiss()
        }
    }

    override fun refreshContent()
    {
        // No-Op for BottomSheetDialog
    }

    override fun isLaunchedInDrillIn(): Boolean
    {
        return false
    }

    override fun isHeaderVisible(): Boolean
    {
        return false
    }

    override fun hasFixedDimensions(): Boolean
    {
        return false
    }

    override fun getDimension(): Point?
    {
        return null
    }

    override fun repositionSameContent()
    {
        // No-Op for BottomSheetDialog
    }

    override fun setViewPortSize(viewPortSize: Point)
    {
        Trace.d(LOG_TAG, "setViewPortSize not implemented for OfficeBottonSheetDialog")
    }

    override fun pushViewProvider(viewProvider: IViewProvider)
    {
        mViewProviderStack.push(viewProvider)
        Trace.i(LOG_TAG, "view provider added to the stack. stack size = " + mViewProviderStack.size)
    }

    override fun showFlyoutContent()
    {
        if (mSurfaceLauncherView != null && !mSurfaceLauncherView!!.isInCheckedState)
        {
            // If showFlyoutContent is called async and the state of the launcher button is not
            // checked then the bottom sheet need not open.
            return
        }
        show()
        mSurfaceLauncherView?.onSurfaceShown()
    }

    override fun hideFlyoutContent()
    {
        // No-Op for BottomSheetDialog
    }

    fun addView(contentView: View)
    {
        mContentView = contentView
    }

    protected fun getViewProvider(dataSource: FlexDataSourceProxy, factory: IControlFactory,
                                  surfaceLauncherView: ISurfaceLauncherView?): IViewProvider
    {
        // This should be changed to take a ViewProvider rather than hard coding
        if (dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSMenuSPProxy.TypeId)
        {
            val fsMenuProxy = FSMenuSPProxy(dataSource)
            var viewProvider: IViewProvider?
            if (fsMenuProxy.getCustomMenuContent())
            {
                viewProvider = CustomViewProviderRegistry.getInstance().getViewProvider(fsMenuProxy.getTcid(), CustomViewProviderRegistry.SurfaceStyle.Menu)
                if (viewProvider != null)
                {
                    return viewProvider
                }
                else
                {
                    viewProvider = CustomViewProviderRegistry.getInstance().getViewProvider(fsMenuProxy.getTcid())
                    if (viewProvider != null)
                    {
                        return viewProvider
                    }
                }
            }
            return createMenuViewProvider(dataSource, factory, surfaceLauncherView)
        }
        else if (dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSSplitMenuSPProxy.TypeId ||
                dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSChunkSPProxy.TypeId ||
                dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSInlineMenuSPProxy.TypeId)
        {
            return createMenuViewProvider(dataSource, factory, surfaceLauncherView)
        }
        else if (dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSImmersiveGallerySPProxy.TypeId)
        {
            val gallerySPProxy = FSImmersiveGallerySPProxy(dataSource)
            val anchorType = AnchorTypes.getAnchorTypeForValue(gallerySPProxy.getAnchorType())
            var viewProvider: IViewProvider?
            if (anchorType == AnchorTypes.Facepile)
            {
                viewProvider = CustomViewProviderRegistry.getInstance().getViewProvider(gallerySPProxy.getTcid())
                if (viewProvider != null)
                {
                    return viewProvider
                }
            }
            viewProvider = FSImmersiveGalleryViewProvider(mContext, dataSource, factory, this /*launchableSurface*/)
            viewProvider.surfaceLauncherView = surfaceLauncherView
            return viewProvider
        }
        else if (dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSColorPickerSPProxy.TypeId)
        {
            if (surfaceLauncherView is MoreColorsButton)
            {
                val viewProvider = MoreColorsViewProvider(mContext, dataSource, factory, this /*launchableSurface*/)
                viewProvider.surfaceLauncherView = surfaceLauncherView
                viewProvider.setMoreColorSelectionObserver(surfaceLauncherView.moreColorSelectionObserver)
                return viewProvider
            }
            val viewProvider = FSColorPickerViewProvider(mContext, dataSource, factory, this /*launchableSurface*/)
            viewProvider.surfaceLauncherView = surfaceLauncherView
            return viewProvider
        }
        else if (dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSComboBoxSPProxy.TypeId)
        {
            val viewProvider = FSComboBoxViewProvider(mContext, dataSource, factory, this /*launchableSurface*/)
            viewProvider.surfaceLauncherView = surfaceLauncherView
            return viewProvider
        }
        else if (dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSColorWheelSPProxy.TypeId)
        {
            val viewProvider = FSColorGridViewProvider(mContext, dataSource, this /*launchableSurface*/)
            viewProvider.surfaceLauncherView = surfaceLauncherView
            return viewProvider
        }
        else if (dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSToolboxSPProxy.TypeId)
        {
            val viewProvider = ToolboxViewProvider(mContext, dataSource, factory, this /*launchableSurface*/)
            viewProvider.surfaceLauncherView = surfaceLauncherView
            return viewProvider
        }
        else if (dataSource.dataSourceDescriptionTypeId == com.microsoft.office.officespace.autogen.FSMoreColorPickerSPProxy.TypeId)
        {
            val viewProvider = MoreColorsViewProvider(mContext, dataSource, factory, this /*launchableSurface*/)
            viewProvider.surfaceLauncherView = surfaceLauncherView
            return viewProvider
        }
        throw IllegalArgumentException("Can't create viewProvider for the given datasource")
    }

    protected fun createMenuViewProvider(dataSource: FlexDataSourceProxy, factory: IControlFactory, surfaceLauncherView: ISurfaceLauncherView?): IViewProvider
    {
        val viewProvider = FSMenuViewProvider(mContext, dataSource, factory, this /*launchableSurface*/)
        // For drill-in floatie surfaceLauncherView could be null
        if (surfaceLauncherView != null)
        {
            viewProvider.setShowHeader(surfaceLauncherView.showHeader)
        }
        return viewProvider
    }

    private fun setBottomSheetContent()
    {
        if (mViewProviderStack.size > 0)
        {
            val viewProvider = mViewProviderStack.peek()
            viewProvider.launchableSurface = this
            if (!viewProvider.isAsyncViewProvider)
            {
                setContentFromViewProvider(viewProvider)
            }
            else
            {
                viewProvider.updateContent { asyncViewProvider ->
                    if (isValidViewProvider(asyncViewProvider))
                    {
                        setContentFromViewProvider(asyncViewProvider)
                    }
                }
            }
        }
        else
        {
            throw IllegalStateException("Neither BottomSheet content view nor BottomSheet data source is set")
        }
    }

    private fun getBottomSheetContainer( ) : LinearLayout
    {
        val container = LinearLayout(mContext)
        container.id = R.id.bottomSheetContainer
        container.orientation = LinearLayout.VERTICAL
        container.addView(getNotchForBottomSheet())
        container.onFocusChangeListener = mFocusChangeListener
        container.isFocusable = true
        container.accessibilityDelegate = mAccessibilityDelegate
        return container
    }

    private fun getNotchForBottomSheet( ) : View
    {
        val notch = ImageView(mContext)
        val imageParams = LinearLayout.LayoutParams(WRAP_CONTENT, WRAP_CONTENT)
        imageParams.gravity = Gravity.CENTER_HORIZONTAL
        notch.layoutParams = imageParams
        notch.setBackgroundResource(R.drawable.sharedux_rounded_dialog_notch)
        return notch
    }

    private fun getLayoutParamsForContainer( contentView: View? ): LinearLayout.LayoutParams
    {
        val layoutParams:LinearLayout.LayoutParams
        if (contentView?.layoutParams != null)
        {
            layoutParams = contentView.layoutParams as LinearLayout.LayoutParams
        }
        else
        {
            layoutParams = LinearLayout.LayoutParams(MATCH_PARENT, MATCH_PARENT)
        }
        layoutParams.setMargins(0 /*left*/,
                0 /*top*/,
                0 /*right*/,
                8 /*bottom*/)
        return layoutParams
    }

    protected fun isValidViewProvider(asyncViewProvider: IViewProvider): Boolean
    {
        return mViewProviderStack.search(asyncViewProvider) == 1
    }

    protected fun setContentFromViewProvider(asyncViewProvider: IViewProvider)
    {
        // move focus to the focus scope since content is getting reset
        val hideHeader = asyncViewProvider.isHeaderHidden
        val contentView = asyncViewProvider.view
        val container = getBottomSheetContainer()
        container.layoutParams = getLayoutParamsForContainer(contentView)
        container.addView(contentView)
        setContentView(container)
        if (!hideHeader)
        {
            setTitle(asyncViewProvider.label)
        }
    }

    private fun expandBottomSheetDialog(hasFocus: Boolean)
    {
        if (hasFocus)
        {
            val bottomSheet = findViewById<FrameLayout>(android.support.design.R.id.design_bottom_sheet)
            val behavior = BottomSheetBehavior.from(bottomSheet)
            behavior.state = BottomSheetBehavior.STATE_EXPANDED
        }
    }

    companion object
    {
        private val LOG_TAG = "OfficeBottomSheetDialog"
        private val PORTRAIT_PEEK_RATIO = 0.4
        private val LANDSCAPE_PEEK_RATIO = 0.66
    }
}